==> invenio_jobs-0.3.3-py2.py3-none-any.whl: invenio_jobs/services/schema.py <==
# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Jobs is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Service schemas."""

import inspect
from datetime import timezone

from invenio_i18n import lazy_gettext as _
from invenio_users_resources.services import schemas as user_schemas
from marshmallow import EXCLUDE, Schema, fields, validate
from marshmallow_oneofschema import OneOfSchema
from marshmallow_utils.fields import SanitizedUnicode, TZDateTime
from marshmallow_utils.permissions import FieldPermissionsMixin
from marshmallow_utils.validators import LazyOneOf

from ..models import RunStatusEnum
from ..proxies import current_jobs


def _not_blank(**kwargs):
    """Returns a non-blank validation rule."""
    max_ = kwargs.get("max", "")
    return validate.Length(
        error=_(
            "Field cannot be blank or longer than {max_} characters.".format(max_=max_)
        ),
        min=1,
        **kwargs,
    )


class TaskParameterSchema(Schema):
    """Schema for a task parameter."""

    name = SanitizedUnicode()

    # TODO: Make custom schema for serializing parameter types
    default = fields.Method("dump_default")
    kind = fields.String()

    def dump_default(self, obj):
        """Dump the default value."""
        if obj.default in (None, inspect.Parameter.empty):
            return None
        elif isinstance(obj.default, (bool, int, float, str)):
            return obj.default
        else:
            return str(obj.default)


class TaskSchema(Schema, FieldPermissionsMixin):
    """Schema for a task."""

    name = SanitizedUnicode()
    description = SanitizedUnicode()
    parameters = fields.Dict(
        keys=SanitizedUnicode(),
        values=fields.Nested(TaskParameterSchema),
    )


class IntervalScheduleSchema(Schema):
    """Schema for an interval schedule based on ``datetime.timedelta``."""

    type = fields.Constant("interval")

    days = fields.Integer()
    seconds = fields.Integer()
    microseconds = fields.Integer()
    milliseconds = fields.Integer()
    minutes = fields.Integer()
    hours = fields.Integer()
    weeks = fields.Integer()


class CrontabScheduleSchema(Schema):
    """Schema for a crontab schedule."""

    type = fields.Constant("crontab")

    minute = fields.String(load_default="*")
    hour = fields.String(load_default="*")
    day_of_week = fields.String(load_default="*")
    day_of_month = fields.String(load_default="*")
    month_of_year = fields.String(load_default="*")


class ScheduleSchema(OneOfSchema):
    """Schema for a schedule."""

    def get_obj_type(self, obj):
        """Get type from object data."""
        if isinstance(obj, dict) and "type" in obj:
            return obj["type"]

    type_schemas = {
        "interval": IntervalScheduleSchema,
        "crontab": CrontabScheduleSchema,
    }
    type_field_remove = False


class JobSchema(Schema, FieldPermissionsMixin):
    """Base schema for a job."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    id = fields.UUID(dump_only=True)

    created = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)
    updated = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)

    title = SanitizedUnicode(required=True, validate=_not_blank(max=250))
    description = SanitizedUnicode()

    active = fields.Boolean(load_default=True)

    task = fields.String(
        required=True,
        validate=LazyOneOf(choices=lambda: current_jobs.tasks.keys()),
    )
    default_queue = fields.String(
        validate=LazyOneOf(choices=lambda: current_jobs.queues.keys()),
        load_default=lambda: current_jobs.default_queue,
    )

    default_args = fields.Dict(
        load_default=dict, allow_none=True, metadata={"type": "json"}
    )

    schedule = fields.Nested(ScheduleSchema, allow_none=True, load_default=None)

    last_run = fields.Nested(lambda: RunSchema, dump_only=True)


class UserSchema(OneOfSchema):
    """User schema."""

    def get_obj_type(self, obj):
        """Get type from object data."""
        return "system" if obj is None else "user"

    type_schemas = {
        "user": user_schemas.UserSchema,
        "system": user_schemas.SystemUserSchema,
    }


class RunSchema(Schema, FieldPermissionsMixin):
    """Base schema for a job run."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    id = fields.UUID(dump_only=True)
    job_id = fields.UUID(dump_only=True)

    created = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)
    updated = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)

    started_by_id = fields.Integer(dump_only=True)
    started_by = fields.Nested(UserSchema, dump_only=True)

    started_at = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)
    finished_at = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)

    status = fields.Enum(RunStatusEnum, dump_only=True)
    message = SanitizedUnicode(dump_only=True)

    task_id = fields.UUID(dump_only=True)

    # Input fields
    title = SanitizedUnicode(validate=_not_blank(max=250))
    args = fields.Dict()
    queue = fields.String(
        validate=LazyOneOf(choices=lambda: current_jobs.queues.keys()),
    )

==> invenio_jobs-3.0.2-py2.py3-none-any.whl: invenio_jobs/services/schema.py <==
# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Jobs is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Service schemas."""

import inspect
import json
from copy import deepcopy
from datetime import timezone

from invenio_i18n import lazy_gettext as _
from invenio_users_resources.services import schemas as user_schemas
from marshmallow import EXCLUDE, Schema, fields, post_load, pre_dump, pre_load, validate
from marshmallow_oneofschema import OneOfSchema
from marshmallow_utils.fields import SanitizedUnicode, TZDateTime
from marshmallow_utils.permissions import FieldPermissionsMixin
from marshmallow_utils.validators import LazyOneOf

from ..models import RunStatusEnum, Task
from ..proxies import current_jobs


def _not_blank(**kwargs):
    """Returns a non-blank validation rule."""
    max_ = kwargs.get("max", "")
    return validate.Length(
        error=_(
            "Field cannot be blank or longer than {max_} characters.".format(max_=max_)
        ),
        min=1,
        **kwargs,
    )


class TaskParameterSchema(Schema):
    """Schema for a task parameter."""

    name = SanitizedUnicode()

    # TODO: Make custom schema for serializing parameter types
    default = fields.Method("dump_default")
    kind = fields.String()

    def dump_default(self, obj):
        """Dump the default value."""
        if obj.default in (None, inspect.Parameter.empty):
            return None
        elif isinstance(obj.default, (bool, int, float, str)):
            return obj.default
        else:
            return str(obj.default)


class TaskSchema(Schema, FieldPermissionsMixin):
    """Schema for a task."""

    name = SanitizedUnicode()
    description = SanitizedUnicode()
    parameters = fields.Dict(
        keys=SanitizedUnicode(),
        values=fields.Nested(TaskParameterSchema),
    )


class IntervalScheduleSchema(Schema):
    """Schema for an interval schedule based on ``datetime.timedelta``."""

    type = fields.Constant("interval")

    days = fields.Integer()
    seconds = fields.Integer()
    microseconds = fields.Integer()
    milliseconds = fields.Integer()
    minutes = fields.Integer()
    hours = fields.Integer()
    weeks = fields.Integer()


class CrontabScheduleSchema(Schema):
    """Schema for a crontab schedule."""

    type = fields.Constant("crontab")

    minute = fields.String(load_default="*")
    hour = fields.String(load_default="*")
    day_of_week = fields.String(load_default="*")
    day_of_month = fields.String(load_default="*")
    month_of_year = fields.String(load_default="*")


class CustomArgsSchema(Schema):
    """Custom arguments schema."""

    args = fields.Raw(load_default=dict, allow_none=True)


class ScheduleSchema(OneOfSchema):
    """Schema for a schedule."""

    def get_obj_type(self, obj):
        """Get type from object data."""
        if isinstance(obj, dict) and "type" in obj:
            return obj["type"]

    type_schemas = {
        "interval": IntervalScheduleSchema,
        "crontab": CrontabScheduleSchema,
    }
    type_field_remove = False


class JobArgumentsSchema(OneOfSchema):
    """Base schema for tasks with arguments."""

    type_field_remove = False
    type_field = "job_arg_schema"

    def __init__(self, *args, **kwargs):
        """Constructor."""
        self.type_schemas = deepcopy(current_jobs.registry.schemas)
        self.type_schemas["custom"] = CustomArgsSchema
        super().__init__(*args, **kwargs)

    def get_obj_type(self, obj):
        """Return object type."""
        if isinstance(obj, dict) and "job_arg_schema" in obj:
            return obj["job_arg_schema"]
        if isinstance(obj, dict) and "job_arg_schema" not in obj:
            return "custom"

    def get_data_type(self, data):
        """Get data type. Defaults to custom if no type is provided."""
        data_type = super().get_data_type(data)
        if data_type is None:
            return "custom"


class JobSchema(Schema, FieldPermissionsMixin):
    """Base schema for a job."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    id = fields.UUID(dump_only=True)

    created = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)
    updated = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)

    title = SanitizedUnicode(required=True, validate=_not_blank(max=250))
    description = SanitizedUnicode()

    active = fields.Boolean(load_default=True)

    task = fields.String(
        required=True,
        validate=LazyOneOf(choices=lambda: [name for name, t in Task.all().items()]),
    )
    default_queue = fields.String(
        validate=LazyOneOf(choices=lambda: current_jobs.queues.keys()),
        load_default=lambda: current_jobs.default_queue,
    )

    default_args = fields.Raw(dump_only=True, dump_default=dict, load_default=dict)

    schedule = fields.Nested(ScheduleSchema, allow_none=True, load_default=None)

    last_run = fields.Nested(lambda: RunSchema, dump_only=True)
    last_runs = fields.Raw(dump_only=True)

    @pre_dump
    def dump_last_runs(self, obj, many=False, **kwargs):
        """Dump last runs of a job."""
        last_runs = obj.get("last_runs", {})
        for key, value in last_runs.items():
            if value:
                last_runs[key] = RunSchema().dump(value.dump())
        return obj


class UserSchema(OneOfSchema):
    """User schema."""

    def get_obj_type(self, obj):
        """Get type from object data."""
        return "system" if obj is None else "user"

    type_schemas = {
        "user": user_schemas.UserSchema,
        "system": user_schemas.SystemUserSchema,
    }


class RunSchema(Schema, FieldPermissionsMixin):
    """Base schema for a job run."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    id = fields.UUID(dump_only=True)
    job_id = fields.UUID(dump_only=True)

    created = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)
    updated = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)

    started_by_id = fields.Integer(dump_only=True)
    started_by = fields.Nested(UserSchema, dump_only=True)

    started_at = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)
    finished_at = TZDateTime(timezone=timezone.utc, format="iso", dump_only=True)

    status = fields.Enum(RunStatusEnum, dump_only=True)
    message = SanitizedUnicode(dump_only=True)

    task_id = fields.UUID(dump_only=True)

    # Input fields
    title = SanitizedUnicode(validate=_not_blank(max=250), dump_default="Manual run")
    args = fields.Nested(
        lambda: JobArgumentsSchema,
        metadata={
            "type": "dynamic",
            "endpoint": "/api/tasks/<item_id>/args",
            "depends_on": "task",
        },
    )
    custom_args = fields.Raw(
        load_default=dict,
        allow_none=True,
        metadata={
            "title": "Custom args",
            "description": "Advanced configuration for seasoned administrators.",
        },
    )
    queue = fields.String(
        validate=LazyOneOf(choices=lambda: current_jobs.queues.keys()),
    )

    @pre_load
    def wrap_args(self, obj, many, **kwargs):
        """Workaround for nested args."""
        # it is possible job has no arguments, that's why we use .get
        obj["args"] = {"args": obj.get("args", {})}
        return obj

    @post_load
    def pick_args(self, obj, many, **kwargs):
        """Choose custom or default args."""
        custom_args = obj.pop("custom_args")
        if custom_args:
            obj["args"] = json.loads(custom_args)
        return obj
